#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#ifdef rtems
#define boolean bfd_boolean_t
#endif
#include "bfd.h"
#include "dis-asm.h"

#ifdef ARCH_a29k
#  define BFD_DISASSEMBLE print_insn_big_a29k
#endif
#ifdef ARCH_alpha
#  define BFD_DISASSEMBLE print_insn_alpha
#endif
#ifdef ARCH_arc
#  define BFD_DISASSEMBLE arc_get_disassembler (bfd_get_mach (abfd), bfd_big_endian (abfd))
#endif
#ifdef ARCH_arm
#  ifdef BFD_ENDIAN_BIG
#    define BFD_DISASSEMBLE print_insn_big_arm
#  else
#    define BFD_DISASSEMBLE print_insn_little_arm
#  endif
#endif
#ifdef ARCH_avr
#  define BFD_DISASSEMBLE print_insn_avr
#endif
#ifdef ARCH_d10v
#  define BFD_DISASSEMBLE print_insn_d10v
#endif
#ifdef ARCH_d30v
#  define BFD_DISASSEMBLE print_insn_d30v
#endif
#ifdef ARCH_h8300
#  define BFD_DISASSEMBLE  ((bfd_get_mach(abfd) == bfd_mach_h8300h)?print_insn_h8300h:((bfd_get_mach(abfd) == bfd_mach_h8300s)?print_insn_h8300s:print_insn_h8300))
#endif
#ifdef ARCH_h8500
#  define BFD_DISASSEMBLE print_insn_h8500
#endif
#ifdef ARCH_hppa
#  define BFD_DISASSEMBLE print_insn_hppa
#endif
#ifdef ARCH_i370
#  define BFD_DISASSEMBLE print_insn_i370
#endif
#ifdef ARCH_i386
#  define BFD_DISASSEMBLE ((bfd_get_mach (abfd) == bfd_mach_i386_i386_intel_syntax)?print_insn_i386_intel:print_insn_i386_att)
#endif
#ifdef ARCH_i960
#  define BFD_DISASSEMBLE print_insn_i960
#endif
#ifdef ARCH_fr30
#  define BFD_DISASSEMBLE print_insn_fr30
#endif
#ifdef ARCH_m32r
#  define BFD_DISASSEMBLE print_insn_m32r
#endif
#ifdef ARCH_m68k
#  define BFD_DISASSEMBLE print_insn_m68k
#endif
#ifdef ARCH_m88k
#  define BFD_DISASSEMBLE print_insn_m88k
#endif
#ifdef ARCH_ns32k
#  define BFD_DISASSEMBLE print_insn_ns32k
#endif
#ifdef ARCH_mcore
#  define BFD_DISASSEMBLE print_insn_mcore
#endif
#ifdef ARCH_mips
#  ifdef BFD_ENDIAN_BIG
#    define BFD_DISASSEMBLE print_insn_big_mips
#  else
#    define BFD_DISASSEMBLE print_insn_little_mips
#  endif
#endif
#ifdef ARCH_mn10200
#  define BFD_DISASSEMBLE print_insn_mn10200
#endif
#ifdef ARCH_mn10300
#  define BFD_DISASSEMBLE print_insn_mn10300
#endif
#ifdef ARCH_pj
#  define BFD_DISASSEMBLE print_insn_pj
#endif
#ifdef ARCH_powerpc
#  ifdef BFD_ENDIAN_BIG
#    define BFD_DISASSEMBLE print_insn_big_powerpc
#  else
#    define BFD_DISASSEMBLE print_insn_little_powerpc
#  endif
#endif
#ifdef ARCH_rs6000
#  define BFD_DISASSEMBLE print_insn_rs6000
#endif
#ifdef ARCH_sh
#  ifdef BFD_ENDIAN_BIG
#    define BFD_DISASSEMBLE print_insn_sh
#  else
#    define BFD_DISASSEMBLE print_insn_shl
#  endif
#endif
#ifdef ARCH_sparc
#  define BFD_DISASSEMBLE print_insn_sparc
#endif
#ifdef ARCH_tic30
#  define BFD_DISASSEMBLE print_insn_tic30
#endif
#ifdef ARCH_tic80
#  define BFD_DISASSEMBLE print_insn_tic80
#endif
#ifdef ARCH_v850
#  define BFD_DISASSEMBLE print_insn_v850
#endif
#ifdef ARCH_w65
#  define BFD_DISASSEMBLE print_insn_w65
#endif
#ifdef ARCH_z8k
#  define BFD_DISASSEMBLE ((bfd_get_mach(abfd) == bfd_mach_z8001)?print_insn_z8001:print_insn_z8002)
#endif
#ifdef ARCH_vax
#  define BFD_DISASSEMBLE print_insn_vax
#endif


static
int read_memory(bfd_vma src, bfd_byte *dst, int length,
                struct disassemble_info *info){
  memcpy((void*)dst,(void*)src,length);
  return 0;
}

static
void memory_error(int status, bfd_vma memaddr, 
                  struct disassemble_info *info){
}

static
void print_address(bfd_vma addr, struct disassemble_info *info){
  info->fprintf_func(info->stream,"0x%x",addr);
}

static
bfd_vma bfd_getb32 (addr)
     register const bfd_byte *addr;
{
  return (((((bfd_vma)addr[0] << 8) | addr[1]) << 8)
          | addr[2]) << 8 | addr[3];
}

static
bfd_vma bfd_getl32 (addr)
     register const bfd_byte *addr;
{
  return (((((bfd_vma)addr[3] << 8) | addr[2]) << 8)
          | addr[1]) << 8 | addr[0];
}


static disassemble_info info;

void debugInit(){
  info.fprintf_func = fprintf;
  info.stream = stdout;
  
  info.flavour =   bfd_target_unknown_flavour;
  info.arch = BFD_ARCH;
  info.mach = BFD_MACH;
  info.endian = BFD_ENDIAN;
  info.symbols = 0;
  info.num_symbols = 0;
  info.flags = 0;
  
  info.read_memory_func = read_memory;
  info.memory_error_func = memory_error;
  info.print_address_func = print_address;
  info.symbol_at_address_func = 0;
  info.buffer = 0;
  info.buffer_vma = 0;
  info.buffer_length = 0;
}


void l(char *addr){
  int i = 0;
  int t;
  static char *last_addr = 100;

  if(addr==0)
    addr=last_addr;

  while(i < 10){
    info.fprintf_func(info.stream,"%8x:\t%02x %02x %02x %02x\t",
                      addr,0xf & addr[0],0xf & addr[1],
                      0xf & addr[2],0xf & addr[3]);
    if((t = BFD_DISASSEMBLE(addr,&info))<0) 
      break;
    info.fprintf_func(info.stream,"\n");
    addr+=t;
    i ++;
  }
  last_addr=addr;
}

 
