%{
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "libld.h"
#include "libstab.h"

#define MAX_ARGS 10

  /* must have MAX_ARGS arguments */
  typedef int (*ff_t)
    (unsigned long,unsigned long,unsigned long,unsigned long,unsigned long,
     unsigned long,unsigned long,unsigned long,unsigned long,unsigned long);



  unsigned long arglist[MAX_ARGS];
  int curr_arg = 0;
  
#define ffcall(p) \
  (((ff_t)p)(arglist[0],arglist[1],arglist[2],arglist[3],arglist[4],\
             arglist[5],arglist[6],arglist[7],arglist[8],arglist[9]))

  void reset_args() {
    memset(arglist, 0, sizeof(arglist));
    curr_arg = 0;
  }

  void add_arg(unsigned long arg){
    if (curr_arg >= MAX_ARGS) return;
    arglist[curr_arg++] = arg;
  }

  int right_of_equal_sign = 0;
  extern char *yytext;

%}

%token IDENTIFIER CONSTANT STRING_LITERAL
%token INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP EQ_OP NE_OP
%token AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN ADD_ASSIGN
%token SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN
%token XOR_ASSIGN OR_ASSIGN 

%start expression_statement
%%

primary_expression
        : IDENTIFIER
{ 
  char *sym = strdup(yytext);
  stab_entry *i = stab_find_by_name(system_symbol_table, sym);

  /* lookup identifier here */ 
  if (i) {
    if ((i->type & N_TEXT) == N_TEXT)
      $$ = (unsigned int) i->address;
    else
      $$ = *(unsigned int*) i->address;
    free(sym);
  }else{
    /* symbol doesn't exist - create it */
    char *var = malloc(sizeof(char*));
    $$ = 0;
    if(var){
      stab_insert(system_symbol_table,sym,var,N_EXT|N_DATA,-1,-1);
      $$ = (unsigned int)var;
      printf("new symbol \"%s\" added to symbol table.\n", sym);
      printf("%s = %d = 0x%x\n", yytext, var, var);
    }
  }
}
        | CONSTANT
{ 
  $$ = $1;
}
        | STRING_LITERAL
{ $$ = $1; }
        | '(' expression ')'
{ $$ = $2; }
        ;

postfix_expression
        : primary_expression
{ $$ = $1; }
        | postfix_expression '[' expression ']'
{ $$ = ((unsigned long *)$1)[$3]; }
        | postfix_expression '(' ')'
{   
  reset_args();
  $$ = ffcall($1);
}
        | postfix_expression '(' argument_expression_list ')'
{ 
  $$ = ffcall($1);
  reset_args();
}
        | postfix_expression argument_expression_list
{ 
  $$ = ffcall($1);
  reset_args();
}
        | postfix_expression INC_OP
{ $$ = $$ + 1; }
        | postfix_expression DEC_OP
{ $$ = $$ - 1; }
        ;

argument_expression_list
        : assignment_expression
{ add_arg($1); }
        | argument_expression_list ',' assignment_expression
{ add_arg($3); }
        ;

unary_expression
        : postfix_expression
{ $$ = $1; }
        | INC_OP unary_expression
{ $$ = $2 + 1; }
        | DEC_OP unary_expression
{ $$ = $2 - 1; }
        | '&' unary_expression
{ $$ = $2; }
        | '*' unary_expression
{ $$ = * (unsigned long *)$2; }
        | '+' unary_expression
{ $$ = + $2; }
        | '-' unary_expression
{ $$ = - $2; }
        | '~' unary_expression
{ $$ = ~ $2; }
        | '!' unary_expression
{ $$ = ! $2; }
        ;

multiplicative_expression
        : unary_expression
{ $$ = $1; }
        | multiplicative_expression '*' unary_expression
{ $$ = $1 * $3; }
        | multiplicative_expression '/' unary_expression
{ $$ = $1 / $3; }
        | multiplicative_expression '%' unary_expression
{ $$ = $1 % $3; }
        ;

additive_expression
        : multiplicative_expression
{ $$ = $1; }
        | additive_expression '+' multiplicative_expression
{ $$ = $1 + $3; }
        | additive_expression '-' multiplicative_expression
{ $$ = $1 - $3; }
        ;

shift_expression
        : additive_expression
{ $$ = $1; }
        | shift_expression LEFT_OP additive_expression
{ $$ = $1 << $3; }
        | shift_expression RIGHT_OP additive_expression
{ $$ = $1 >> $3; }
        ;

relational_expression
        : shift_expression
{ $$ = $1; }
        | relational_expression '<' shift_expression
{ $$ = $1 < $3; }
        | relational_expression '>' shift_expression
{ $$ = $1 > $3; }
        | relational_expression LE_OP shift_expression
{ $$ = $1 <= $3; }
        | relational_expression GE_OP shift_expression
{ $$ = $1 >= $3; }
        ;

equality_expression
        : relational_expression
{ $$ = $1; }
        | equality_expression EQ_OP relational_expression
{ $$ = $1 == $3; }
        | equality_expression NE_OP relational_expression
{ $$ = $1 != $3; }
        ;

and_expression
        : equality_expression
{ $$ = $1; }
        | and_expression '&' equality_expression
{ $$ = $1 & $3; }
        ;

exclusive_or_expression
        : and_expression
{ $$ = $1; }
        | exclusive_or_expression '^' and_expression
{ $$ = $1 ^ $3; }
        ;

inclusive_or_expression
        : exclusive_or_expression
{ $$ = $1; }
        | inclusive_or_expression '|' exclusive_or_expression
{ $$ = $1 | $3; }
        ;

logical_and_expression
        : inclusive_or_expression
{ $$ = $1; }
        | logical_and_expression AND_OP inclusive_or_expression
{ $$ = $1 && $3; }
        ;

logical_or_expression
        : logical_and_expression
{ $$ = $1; }
        | logical_or_expression OR_OP logical_and_expression
{ $$ = $1 || $3; }
        ;

conditional_expression
        : logical_or_expression
{ $$ = $1; }
        | logical_or_expression '?' expression ':' conditional_expression
{ $$ = $1 ? $3 : $5; }
        ;

assignment_expression
        : conditional_expression
{ $$ = $1; }
        | unary_expression '=' { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$$ = $4; right_of_equal_sign = 0; }
        | unary_expression MUL_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 *= $4; right_of_equal_sign = 0; }
        | unary_expression DIV_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 /= $4; right_of_equal_sign = 0; }
        | unary_expression MOD_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 %= $4; right_of_equal_sign = 0; }
        | unary_expression ADD_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 += $4; right_of_equal_sign = 0; }
        | unary_expression SUB_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 -= $4; right_of_equal_sign = 0; }
        | unary_expression LEFT_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 <<= $4; right_of_equal_sign = 0; }
        | unary_expression RIGHT_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 >>= $4; right_of_equal_sign = 0; }
        | unary_expression AND_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 &= $4; right_of_equal_sign = 0; }
        | unary_expression XOR_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 ^= $4; right_of_equal_sign = 0; }
        | unary_expression OR_ASSIGN { right_of_equal_sign = 1; } assignment_expression
{ $$ = *(unsigned long*)$1 |= $4; right_of_equal_sign = 0; }
        ;

expression
        : assignment_expression
{ $$ = $1; }
        | expression ',' assignment_expression
{ $$ = $3; }
        ;

expression_statement
        : '\n'
{ /* no action */ return 1; }
        | ';' '\n'
{ /* no action */ return 1; }
        | expression '\n'
{ 
  printf("value %d = 0x%x\n\n", $1, $1); 
  return 1;
}
        | expression ';' '\n'
{ 
  printf("value %d = 0x%x\n\n", $1, $1); 
  return 1;
}
        ;


%%
#include <stdio.h>

extern int column;

yyerror(s)
char *s;
{
        fflush(stdout);
        printf("\n%*s\n%*s\n", column, "^", column, s);
}

