
D                       [0-9]
L                       [a-zA-Z_]
H                       [a-fA-F0-9]
E                       [Ee][+-]?{D}+
FS                      (f|F|l|L)
IS                      (u|U|l|L)*

%{
#include <stdio.h>
#include "y.tab.h"

void count();
char *string_cvt(char *);
extern int yylval;
%}

%%
#.*                     { /* ignore comments starting with a '#' */ }
"/*"                    { comment(); }

{L}({L}|{D})*           { count(); return(check_type()); }

0[xX]{H}+{IS}?          { count(); sscanf(yytext,"%x",&yylval); return(CONSTANT); }
0{D}+{IS}?              { count(); sscanf(yytext,"%o",&yylval); return(CONSTANT); }
{D}+{IS}?               { count(); sscanf(yytext,"%d",&yylval); return(CONSTANT); }
'(\\.|[^\\'])+'         { count(); return(CONSTANT); }

{D}+{E}{FS}?            { count(); return(CONSTANT); }
{D}*"."{D}+({E})?{FS}?  { count(); return(CONSTANT); }
{D}+"."{D}*({E})?{FS}?  { count(); return(CONSTANT); }

\"(\\.|[^\\"])*\"       { count(); yylval = (unsigned int)string_cvt(yytext); return(STRING_LITERAL); }

">>="                   { count(); return(RIGHT_ASSIGN); }
"<<="                   { count(); return(LEFT_ASSIGN); }
"+="                    { count(); return(ADD_ASSIGN); }
"-="                    { count(); return(SUB_ASSIGN); }
"*="                    { count(); return(MUL_ASSIGN); }
"/="                    { count(); return(DIV_ASSIGN); }
"%="                    { count(); return(MOD_ASSIGN); }
"&="                    { count(); return(AND_ASSIGN); }
"^="                    { count(); return(XOR_ASSIGN); }
"|="                    { count(); return(OR_ASSIGN); }
">>"                    { count(); return(RIGHT_OP); }
"<<"                    { count(); return(LEFT_OP); }
"++"                    { count(); return(INC_OP); }
"--"                    { count(); return(DEC_OP); }
"&&"                    { count(); return(AND_OP); }
"||"                    { count(); return(OR_OP); }
"<="                    { count(); return(LE_OP); }
">="                    { count(); return(GE_OP); }
"=="                    { count(); return(EQ_OP); }
"!="                    { count(); return(NE_OP); }
";"                     { count(); return(';'); }
"{"                     { count(); return('{'); }
"}"                     { count(); return('}'); }
","                     { count(); return(','); }
":"                     { count(); return(':'); }
"="                     { count(); return('='); }
"("                     { count(); return('('); }
")"                     { count(); return(')'); }
"["                     { count(); return('['); }
"]"                     { count(); return(']'); }
"."                     { count(); return('.'); }
"&"                     { count(); return('&'); }
"!"                     { count(); return('!'); }
"~"                     { count(); return('~'); }
"-"                     { count(); return('-'); }
"+"                     { count(); return('+'); }
"*"                     { count(); return('*'); }
"/"                     { count(); return('/'); }
"%"                     { count(); return('%'); }
"<"                     { count(); return('<'); }
">"                     { count(); return('>'); }
"^"                     { count(); return('^'); }
"|"                     { count(); return('|'); }
"?"                     { count(); return('?'); }
\n                      { count(); return('\n'); }

[ \t\v\f]               { count(); }
.                       { /* ignore bad characters */ }

%%

int yywrap() {
  return 1;
}

comment()
{
        char c, c1;

loop:
        while ((c = input()) != '*' && c != 0)
                putchar(c);

        if ((c1 = input()) != '/' && c != 0)
        {
                unput(c1);
                goto loop;
        }

        if (c != 0)
                putchar(c1);
}


int column = 0;

void count()
{
        int i;

        for (i = 0; yytext[i] != '\0'; i++)
                if (yytext[i] == '\n')
                        column = 0;
                else if (yytext[i] == '\t')
                        column += 8 - (column % 8);
                else
                        column++;

}


int check_type()
{
return(IDENTIFIER);
}


/* convert special characters to their real meanings in a string */
char *string_cvt(char *buf){
  char *s, *d;
  char *rval = malloc(strlen(buf) + 1);
  s=buf;
  d=rval;
  s++;
  for(; *s != '"'; s++, d++){
      *d=*s;
      if (*s == '\\'){
        s++;
	switch(*s){
	case 'a':
  	  *d=0x07;
	  break;
	case 'b':
	  *d=0x08;
          break;
        case 'f':
          *d=0x0c;
          break;
        case 'n':
          *d=0x0a;
	  break;
        case 'r':
          *d=0x0d;
	  break;
        case 't':
          *d=0x09;
	  break;
        case 'v':
          *d=0x0b;
	  break;
        default:
          *d=*s;
	  break;
        }
     }
  }
  *d=0;
  return rval;
}

