/*
 *
 * Copyright 1995-1998 by Metro Link, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Metro Link, Inc. not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Metro Link, Inc. makes no
 * representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 * METRO LINK, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL METRO LINK, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#if defined(UseMMAP) || (defined(linux) && defined(__ia64__))
#include <sys/mman.h>
#endif
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#if defined(linux) && \
    (defined(__alpha__) || defined(__powerpc__) || defined(__ia64__))
#include <malloc.h>
#endif
#include <stdarg.h>
#include "libld.h"
#include "elf.h"
#include "coff.h"

#include "sym.h"
#include "loader.h"
#include "elfloader.h"
#include "aoutloader.h"
#include "coffloader.h"

#ifdef AOUT_SUPPORT
#define LoadModule         AOUTLoadModule
#define ResolveSymbols     AOUTResolveSymbols
#define CheckForUnresolved AOUTCheckForUnresolved
#define AddressToSection   AOUTAddressToSection
#define UnloadModule       AOUTUnloadModule
#endif /* AOUT_SUPPORT */

#ifdef COFF_SUPPORT
#define LoadModule         COFFLoadModule
#define ResolveSymbols     COFFResolveSymbols
#define CheckForUnresolved COFFCheckForUnresolved
#define AddressToSection   COFFAddressToSection
#define UnloadModule       COFFUnloadModule
#endif /* COFF_SUPPORT */

#ifdef ELF_SUPPORT
#define LoadModule         ELFLoadModule
#define ResolveSymbols     ELFResolveSymbols
#define CheckForUnresolved ELFCheckForUnresolved
#define AddressToSection   ELFAddressToSection
#define UnloadModule       ELFUnloadModule
#endif /* ELF_SUPPORT */

/*
#define DEBUG
#define DEBUGAR
#define DEBUGLIST
#define DEBUGMEM
*/

int check_unresolved_sema = 0;

#if defined(Lynx) && defined(sun)
/* Cross build machine doesn;t have strerror() */
#define strerror(err) "strerror unsupported"
#endif

#ifdef __EMX__
void * os2ldcalloc(size_t,size_t);
#endif

/*
 * handles are used to identify files that are loaded. Even archives
 * are counted as a single file.
 */
#define MAX_HANDLE 256
#define HANDLE_FREE 0
#define HANDLE_USED 1
static char freeHandles[MAX_HANDLE] ;
static int refCount[MAX_HANDLE] ;

#if defined(__sparc__) && defined(__GNUC__)
# define SYMFUNCDOT(func) { "." #func, (funcptr)&__sparc_dot_ ## func },
# define SYMFUNCDOT89(func) { "." #func, (funcptr)&func ## _sparcv89 },
# define DEFFUNCDOT(func) 					\
extern void __sparc_dot_ ## func (void) __asm__ ("." #func);	\
extern void func ## _sparcv89 (void);
DEFFUNCDOT(rem)
DEFFUNCDOT(urem)
DEFFUNCDOT(mul)
DEFFUNCDOT(umul)
DEFFUNCDOT(div)
DEFFUNCDOT(udiv)
static LOOKUP SparcV89LookupTab[] = {
   SYMFUNCDOT89(rem)
   SYMFUNCDOT89(urem)
   SYMFUNCDOT89(mul)
   SYMFUNCDOT89(umul)
   SYMFUNCDOT89(div)
   SYMFUNCDOT89(udiv)
   { 0, 0 }
};
static LOOKUP SparcLookupTab[] = {
   SYMFUNCDOT(rem)
   SYMFUNCDOT(urem)
   SYMFUNCDOT(mul)
   SYMFUNCDOT(umul)
   SYMFUNCDOT(div)
   SYMFUNCDOT(udiv)
   { 0, 0 }
};
#ifdef linux
#if defined(__GNUC__) && defined(__GLIBC__)
#define HWCAP_SPARC_MULDIV	8
extern unsigned long int _dl_hwcap;
#endif

static int
sparcUseHWMulDiv(void)
{
    FILE *f;
    char buffer[1024];
    char *p;
#if defined(__GNUC__) && defined(__GLIBC__)
    unsigned long *hwcap;
    __asm(".weak _dl_hwcap");
    
    hwcap = &_dl_hwcap;
    __asm("" : "=r" (hwcap) : "0" (hwcap));
    if (hwcap) {
        if (*hwcap & HWCAP_SPARC_MULDIV)
    	    return 1;
    	else
    	    return 0;
    }
#endif
    f = fopen("/proc/cpuinfo","r");
    if (!f) return 0;
    while (fgets(buffer, 1024, f) != NULL) {
        if (!strncmp (buffer, "type", 4)) {
            p = strstr (buffer, "sun4");
            if (p && (p[4] == 'u' || p[4] == 'd' || p[4] == 'm')) {
                fclose(f);
                return 1;
            }
        }
    }
    fclose(f);
    return 0;
}
#endif
#endif


/*
 * modules are used to identify compilation units (ie object modules).
 * Archives contain multiple modules, each of which is treated seperately.
 */
int moduleseq = 0;

/*
 * GDB Interface
 * =============
 *
 * Linked list of loaded modules - gdb will traverse this to determine
 * whether it needs to add the symbols for the loaded module.
 */
LDRModulePtr ModList = 0;

/* Flag which gdb sets to let us know we're being debugged */
char DebuggerPresent = 0;

/* List of common symbols */
LDRCommonPtr ldrCommons;
int nCommons;

static int fatalReqSym = 0;

/* Prototypes for static functions. */
static loaderPtr _LoaderListPop(int);



void loader_init(void)
{
#ifdef __sparc__
#ifdef linux
    if (sparcUseHWMulDiv())
	LoaderAddSymbols(-1, -1, SparcV89LookupTab ) ;
    else
#endif
	LoaderAddSymbols(-1, -1, SparcLookupTab ) ;
#endif 

#if defined(linux) && \
    (defined(__alpha__) || defined(__powerpc__) || defined(__ia64__))
    /*
     * The glibc malloc uses mmap for large allocations anyway. This breaks
     * some relocation types because the offset overflow. See loader.h for more
     * details. We need to turn off this behavior here.
     */
    mallopt(M_MMAP_MAX,0);
#endif
}

/*
 * Determine what type of object is being loaded.
 * This function is responsible for restoring the offset.
 * The fd and offset are used here so that when Archive processing
 * is enabled, individual elements of an archive can be evaluated
 * so the correct loader_funcs can be determined.
 */
int
_GetModuleType(int fd, long offset)
{
    unsigned char	buf[256]; /* long enough for the largest magic type */

    if( read(fd,buf,sizeof(buf)) < 0 ) {
	return -1;
    }

#ifdef DEBUG
    ErrorF("Checking module type %10s\n", buf );
    ErrorF("Checking module type %x %x %x %x\n", buf[0], buf[1], buf[2], buf[3] );
#endif

    lseek(fd,offset,SEEK_SET);

#if defined(AIAMAG)
    /* LynxOS PPC style archives */
    if (strncmp((char *) buf, AIAMAG, SAIAMAG) == 0) {
	return LD_ARCHIVE;
    }
#endif

    if (strncmp((char *) buf, ELFMAG, SELFMAG) == 0) {
	if( buf[ELFDLOFF] == ELFDLMAG ) {
	    return LD_ELFDLOBJECT;
	} else {
	    return LD_ELFOBJECT;
	}
    }

    if( buf[0] == 0x4c && buf[1] == 0x01 ) {
	/* I386MAGIC */
	return LD_COFFOBJECT;
    }
    if( buf[0] == 0x01 && buf[1] == 0xdf ) {
	/* XCOFFMAGIC */
	return LD_COFFOBJECT;
	}
    if( buf[0] == 0x0d && buf[1] == 0x01 ) {
	/* ZCOFFMAGIC (LynxOS) */
	return LD_COFFOBJECT;
    }
    if( buf[0] == 0x00 && buf[1] == 0x86 && buf[2] == 0x01 && buf[3] == 0x07) {
        /* AOUTMAGIC */
        return LD_AOUTOBJECT;
    }
    if (buf[0] == 0x07 && buf[1] == 0x01 && (buf[2] == 0x64 || buf[2] == 0x86))
    {
        /* AOUTMAGIC, (Linux OMAGIC, old impure format, also used by OS/2 */
        return LD_AOUTOBJECT;
    }
    if (buf[0] == 0x07 && buf[1] == 0x01 && buf[2] == 0x00 && buf[3] == 0x00)
    {
        /* AOUTMAGIC, BSDI */
        return LD_AOUTOBJECT;
    }
    if ((buf[0] == 0xc0 && buf[1] == 0x86) || /* big endian form */
	(buf[3] == 0xc0 && buf[2] == 0x86)) { /* little endian form */
        /* i386 shared object */
        return LD_AOUTDLOBJECT;
    }

    return LD_UNKNOWN;
}


int	offsetbias=0; /* offset into archive */
/*
 * _LoaderFileToMem() loads the contents of a file into memory using
 * the most efficient method for a platform.
 */
void *
_LoaderFileToMem(int fd, unsigned long offset,int size, char *label)
{
  char *ptr;
#ifdef DEBUGMEM
  ErrorF("_LoaderFileToMem(%d,%u(%u),%d,%s)",fd,offset,offsetbias,size,label);
#endif /* DEBUGMEM */
  
  if(size == 0){
#ifdef DEBUGMEM
    ErrorF("=NULL\n",ptr);
#endif /* DEBUGMEM */
    return NULL;
  }

#ifndef __EMX__
  if( (ptr=calloc(size,1)) == NULL )
    FatalError("_LoaderFileToMem() malloc failed\n" );
#else
  if( (ptr=os2ldcalloc(size,1)) == NULL )
    FatalError("_LoaderFileToMem() malloc failed\n" );
#endif /* __EMX__ */

#if defined(linux) && defined(__ia64__)
    {
	unsigned long page_size = getpagesize();
	unsigned long round;

	round = (unsigned long)ptr & (page_size-1);
	mprotect(ptr - round, (size+round+page_size-1) & ~(page_size-1),
		 PROT_READ|PROT_WRITE|PROT_EXEC);
    }
#endif /* linux && __ia64__ */

    if(lseek(fd,offset+offsetbias,SEEK_SET)<0)
	FatalError2("\n_LoaderFileToMem() lseek() failed: %s\n",strerror(errno));

    if(read(fd,ptr,size)!=size)
	FatalError2("\n_LoaderFileToMem() read() failed: %s\n",strerror(errno));

#if defined(linux) && defined(__powerpc__) 
    /*
     * Keep the instruction cache in sync with changes in the
     * main memory.
     */
    { 
	int i; 
	for (i = 0; i < size; i += 16) 
	    ppc_flush_icache(ptr+i); 
	ppc_flush_icache(ptr+size-1); 
    } 
#endif /* linux && __powerpc__ */

#ifdef DEBUGMEM
    ErrorF("=%lx\n",ptr);
#endif /* DEBUGMEM */

    return (void *)ptr;
}

/*
 * _LoaderFreeFileMem() free the memory in which a file was loaded.
 */
void
_LoaderFreeFileMem(void *addr, int size)
{
#ifdef DEBUGMEM
    ErrorF("_LoaderFreeFileMem(%x,%d)\n",addr,size);
#endif
#ifdef UseMMAP
    munmap(addr,size);
#else
    if(size == 0)
	return;

    free(addr);
#endif

    return;
}

int
_LoaderFileRead(int fd, unsigned int offset, void *buf, int size)
{
    if(lseek(fd,offset+offsetbias,SEEK_SET)<0)
	FatalError2("_LoaderFileRead() lseek() failed: %s\n", strerror(errno) );

    if(read(fd,buf,size)!=size)
	FatalError2("_LoaderFileRead() read() failed: %s\n", strerror(errno) );

    return size;
}

static loaderPtr listHead = (loaderPtr) 0 ;

loaderPtr
_LoaderListPush()
{
  loaderPtr item = calloc(1, sizeof (struct _loader));
  item->next = listHead ;
  listHead = item;

  return item;
}

static loaderPtr
_LoaderListPop(int handle)
{
  loaderPtr item=listHead;
  loaderPtr *bptr=&listHead; /* pointer to previous node */

  while(item) {
	if( item->handle == handle ) {
	    *bptr=item->next;	/* remove this from the list */
	    return item;
	    }
	bptr=&(item->next);
	item=item->next;
	}

  return 0;
}

void loader_show() {
  loaderPtr item=listHead;

  printf("Module ID:           Name:\n");
  while(item) {
    printf("%-20s %d\n",item->name,item->handle);
  }
}

/*
 * _LoaderHandleToName() will return the name of the first module with a
 * given handle. This requires getting the last module on the LIFO with
 * the given handle.
 */
char *
_LoaderHandleToName(int handle)
{
  loaderPtr item=listHead;
  loaderPtr aritem=NULL;
  loaderPtr lastitem=NULL;

  if ( handle < 0 ) {
	return "(built-in)";
	}
  while(item) {
	if( item->handle == handle ) {
	    if( strchr(item->name,':') == NULL )
		aritem=item;
	    else
		lastitem=item;
	    }
	item=item->next;
	}

  if( aritem )
    return aritem->name;

  if( lastitem )
    return lastitem->name;

  return 0;
}

/*
 * _LoaderHandleToCanonicalName() will return the cname of the first module
 * with a given handle. This requires getting the last module on the LIFO with
 * the given handle.
 */
char *
_LoaderHandleToCanonicalName(int handle)
{
  loaderPtr item=listHead;
  loaderPtr lastitem=NULL;

  if ( handle < 0 ) {
	return "(built-in)";
  }
  while(item) {
	if( item->handle == handle ) {
		lastitem=item;
	}
	item=item->next;
  }

  if( lastitem )
    return lastitem->cname;

  return NULL;
}

/*
 * _LoaderModuleToName() will return the name of the first module with a
 * given handle. This requires getting the last module on the LIFO with
 * the given handle.
 */
char *
_LoaderModuleToName(int module)
{
  loaderPtr item=listHead;
  loaderPtr aritem=NULL;
  loaderPtr lastitem=NULL;

  if ( module < 0 ) {
	return "(built-in)";
  }
  while(item) {
	if( item->module == module ) {
	    if( strchr(item->name,':') == NULL )
		aritem=item;
	    else
		lastitem=item;
	}
	item=item->next;
  }

  if( aritem )
    return aritem->name;

  if( lastitem )
    return lastitem->name;

  return 0;
}

/*
 * _LoaderAddressToSection() will return the name of the file & section
 * that contains the given address.
 */
int
_LoaderAddressToSection(const unsigned long address, const char **module,
			const char ** section)
{
  loaderPtr item=listHead;

  while(item) {
	if( (*section=AddressToSection(item->private, address)) != NULL ) {
		*module=_LoaderModuleToName(item->module);
		return 1;
	}
	item=item->next;
  }

  return 0;
}


/* 
 * _LoaderHandleUnresolved() decides what to do with an unresolved
 * symbol.  Symbols that are not on the "referenced" or "required" lists
 * get a warning if they are unresolved.  Symbols that are on the "required"
 * list generate a fatal error if they are unresolved.
 */

int
_LoaderHandleUnresolved(char *symbol, char *module)
{
    int fatalsym = 0;

    return(fatalsym);
}

/*
 * Relocation list manipulation routines
 */

/*
 * _LoaderGetRelocations() Return the list of outstanding relocations
 */
LoaderRelocPtr
_LoaderGetRelocations(void *mod)
{
	loader_funcs	*formatrec = (loader_funcs *)mod;

	return  &(formatrec->pRelocs);
}

/*
 * Public Interface to the loader.
 */

module_t loader_load(const char *module)
{
    loaderPtr tmp ;
    module_t new_handle;
    int modtype;
    int fd;
    LOOKUP *pLookup;

#if defined(DEBUG)
    ErrorF("loader_load(%s)\n", module );
#endif

    /*
     * OK, it's a new one. Add it.
     */
#ifdef DEBUG
    ErrorF( "Loading %s\n", module ) ;
#endif

    /*
     * Find a free handle.
     */
    new_handle = 1;
    while ( freeHandles[new_handle] && new_handle < MAX_HANDLE )
	new_handle ++ ;

    if ( new_handle == MAX_HANDLE ) {
	ErrorF( "Out of loader space\n" ) ; /* XXX */
	return -1 ;
    }

    freeHandles[new_handle] = HANDLE_USED ;
    refCount[new_handle] = 1;

    if( (fd=open(module, O_RDONLY)) < 0 ) {
	ErrorF2( "Unable to open %s\n", module );
	freeHandles[new_handle] = HANDLE_FREE ;
	return -1 ;
    }

    if( (modtype=_GetModuleType(fd,0)) < 0 ) {
	ErrorF2( "%s is an unrecognized module type\n", module ) ;
        freeHandles[new_handle] = HANDLE_FREE ;
	return -1;
    }

    tmp=_LoaderListPush();
    tmp->name = malloc(strlen(module) + 1);
    strcpy(tmp->name, module);
    tmp->cname = malloc(strlen(module) + 1);
    strcpy(tmp->cname, module);
    tmp->handle = new_handle;
    tmp->module = moduleseq++;
    tmp->funcs=&funcs;

    if((tmp->private = LoadModule(tmp,fd, &pLookup)) == NULL) {
	ErrorF2( "Failed to load %s\n", module ) ;
	_LoaderListPop(new_handle);
        freeHandles[new_handle] = HANDLE_FREE ;
	return -1;
    }

    LoaderAddSymbols(new_handle, tmp->module, pLookup);
    free(pLookup);

    close(fd);

    return new_handle;
}

void *
LoaderSymbol(const char *sym)
{
  stab_entry * item = NULL;

  ResolveSymbols(&funcs);

  item = stab_find_by_name(system_symbol_table, sym);

  if ( item )
    return item->address ;
  else
#ifdef DLOPEN_SUPPORT
    return(DLFindSymbol(sym));
#else
    return NULL;
#endif
}

int
LoaderResolveSymbols(void)
{
  ResolveSymbols(&funcs);
  return 0;
}

void
Trap()
{
}

void
LoaderDefaultFunc(void)
{
	ErrorF("\n\n\tThis should not happen!\n"
		"\tAn unresolved function was called!\n");

	Trap();
	
	FatalError("\n");
}

int loader_unload(module_t handle)
{
  loaderRec fakeHead ;
  loaderPtr tmp = & fakeHead ;

  if ( handle < 0 || handle > MAX_HANDLE )
	return -1;

 /*
  * check the reference count, only free it if it goes to zero
  */
  if (--refCount[handle])
	return 0;
 /*
  * find the loaderRecs associated with this handle.
  */

  while( (tmp=_LoaderListPop(handle)) != NULL ) {
	if( strchr(tmp->name,':') == NULL ) {
		/* It is not a member of an archive */
		ErrorF2( "Unloading %s\n", tmp->name ) ;
	}
	UnloadModule(tmp->private);
	free(tmp->name);
	free(tmp->cname);
	free(tmp);
  }
  
  freeHandles[handle] = HANDLE_FREE ;

return 0;
}

void
LoaderDuplicateSymbol(const char *symbol, const int handle)
{
    ErrorF3("Duplicate symbol %s in %s\n", symbol,
		listHead ? listHead->name : "(built-in)");
    ErrorF2("Also defined in %s\n", _LoaderHandleToName(handle));
}

/* GDB Sync function */
void _loader_debug_state()
{
}

void
LoaderAddSymbols(handle, module, list)
int	handle;
int	module;
LOOKUP	*list ;
{
    LOOKUP	*l = list;

    if (!list)
	return;
    /* Visit every symbol in the lookup table,
     * and add it to the given namespace.
     */
    while ( l->symName ) {
	stab_insert( system_symbol_table,  
		     l->symName, 
		     l->offset,
		     l->type, 
		     handle,
		     module);
	l ++ ;
    }
}


