#include "loader.h"
#include "elfloader.h"
#include "aoutloader.h"
#include "coffloader.h"

/*
 * Array containing entry points for different formats.
 */
#ifdef AOUT_SUPPORT
#define LoadModule         AOUTLoadModule
#define ResolveSymbols     AOUTResolveSymbols
#define CheckForUnresolved AOUTCheckForUnresolved
#define AddressToSection   AOUTAddressToSection
#define UnloadModule       AOUTUnloadModule
#endif /* AOUT_SUPPORT */

#ifdef COFF_SUPPORT
#define LoadModule         COFFLoadModule
#define ResolveSymbols     COFFResolveSymbols
#define CheckForUnresolved COFFCheckForUnresolved
#define AddressToSection   COFFAddressToSection
#define UnloadModule       COFFUnloadModule
#endif /* COFF_SUPPORT */

#ifdef ELF_SUPPORT
#define LoadModule         ELFLoadModule
#define ResolveSymbols     ELFResolveSymbols
#define CheckForUnresolved ELFCheckForUnresolved
#define AddressToSection   ELFAddressToSection
#define UnloadModule       ELFUnloadModule
#endif /* ELF_SUPPORT */


loader_funcs funcs = {
  LoadModule,
  ResolveSymbols,
  CheckForUnresolved,
  AddressToSection,
  UnloadModule, {0,0,0,0,0}
};
