/*****************************************************************
*
* File:      $Id: libstab.h,v 1.1.1.1 2002/08/02 01:04:47 joel Exp $
* Version:   $Version$
* Author:    G. Cornell
*
*****************************************************************
*
* $Log: libstab.h,v $
* Revision 1.1.1.1  2002/08/02 01:04:47  joel
* base RTEMS shell and dynamic loader
*
* Revision 1.1  2002/06/14 21:37:11  cornell
* Initial revision
*
*
*
*****************************************************************/

/* borrowed from the xfree86 loader, submitted by Metro Link, Inc.: */
/*
 * Copyright 1997,1998 Metro Link, Inc.
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Metro Link, Inc. not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Metro Link, Inc. makes no
 * representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 * METRO LINK, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL METRO LINK, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef LIBSTAB_H_INCLUDED
#define LIBSTAB_H_INCLUDED

/*
 * Symbol types used in n_type below.
 * COFF Section numbers
 */
#define N_TEXT       1
#define N_DATA       2
#define N_BSS        3
#define N_COMMENT    4
#define N_EXT     0x08

#define DEFAULT_MODULE_ID 0

typedef struct _stab_entry {
  char	*name ;
  void	*address ;
  int	type ; /* N_TEXT or N_DATA */
  struct _stab_entry *next ;
  int	handle ;
  int	module ;
#if defined(__powerpc__)
  /*
   * PowerPC file formats require special routines in some circumstances
   * to assist in the linking process. See the specific loader for
   * more details.
   */
  union {
    unsigned short	plt[8];		/* ELF */
    unsigned short	glink[14];	/* XCOFF */
  } code ;
#endif
}stab_entry;

typedef struct {
  int          num_buckets;
  int          multiple_symbols_allowed;
  stab_entry **buckets;
}stab_t;

typedef int (*stab_proc)(stab_t *stab, 
			 stab_entry *sym, 
			 unsigned long user_data);

extern stab_t *system_symbol_table;

/*
 * This is defined when building the system symbol table using
 * the mkstab utility.  This prevents the function prototypes
 * from being declared.
 */
#ifndef NO_FUNCTION_PROTOTYPES

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  stab_t     *stab_create(int num_buckets);
  void        stab_destroy(stab_t *stab);
  int         stab_insert(stab_t        *stab,
			  char	*name,
			  void	*address,
			  int	type,
			  int	handle,
			  int	module);
  int         stab_erase(stab_t     *stab,
			 stab_entry *sym);
  stab_entry *stab_for_each(stab_t        *stab,
			    stab_proc      proc,
			    unsigned long  user_data);
  stab_entry *stab_find_by_name(stab_t      *stab,
				char const  * const name);
  stab_entry *stab_find_by_addr(stab_t  *stab,
				void    *value);

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */

#endif /* !NO_FUNCTION_PROTOTYPES */

#endif /* LIBSTAB_H_INCLUDED */  /* DO NOT EDIT AFTER THIS LINE */
