#################################################################
#
# File:      $Id: mkstab.awk,v 1.3 2002/09/19 01:07:17 glenco Exp $
# Version:   $Version$
# Author:    G. Cornell
#
#################################################################
#
# $Log: mkstab.awk,v $
# Revision 1.3  2002/09/19 01:07:17  glenco
# General cleanup of the code.  The most important changes are in the mkstab utility
#
# Revision 1.2  2002/09/16 00:32:31  glenco
# Adding example program and improving mkstab
#
# Revision 1.3  2002/09/14 19:18:41  glen
# more cleanups
#
# Revision 1.2  2002/09/06 03:06:57  glen
# Cleanup
#
# Revision 1.1.1.1  2002/08/29 02:48:57  glen
# initial import into CVS
#
#
#
#################################################################

# 
# Create a symbol table from the given object file in c source 
# form in order to reduce start-up performace costs, which can be
# significant for large symbol tables.

# USAGE:
# awk -f mkstab
#      [-v system_symbol_table=<name>]
#      [-v num_buckets=<hash_bucket_size>]
#      [-v excludes=<filename>]
#      [-v CPU=i[3456]86|ppc]
#      [-v OMF=elf|aout|coff]

BEGIN {
   # all of the strings are stored in a big array of text:
   strtab_index = 0;
   
   num_entries = 0;

   if(length(system_symbol_table)==0)
     system_symbol_table="system_symbol_table";

   # The stab array is the collection of indecies into the 
   # system table parrallel arrays.
   if (num_buckets <= 0)
	num_buckets = 512;
   for(i = 0; i<num_buckets; i++)
     buckets[i] = -1;

}


function is_weird_symbol(sym){
  if(length(sym)==0)
    return 1;
  if(match(sym,"[^a-zA-Z0-9_]")>0)
    return 1;
  if(index(sym,"_")==1)
    return 1;
  if(sym=="main")
    return 1;
  if(sym=="L1")
    return 1;
  if(sym=="Letext")
    return 1;
  if(sym=="tzstring_list")
    return 1;
  if(sym=="stab_create")
    return 1;
  if(length(excludes)>0)
    return system("egrep \"^" sym "$\" " excludes " >/dev/null 2>&1")==0;

  return 0;
}

function is_text(type){
  return (type == "t") || (type == "T");
}

function get_type(type){
    if(type=="t")
	return "N_TEXT";
    if(type=="T")
	return "N_TEXT|N_EXT";
    if(type=="d")
	return "N_DATA";
    if(type=="D")
	return "N_DATA|N_EXT";
    return "BAD";
}

function print_prototype(i){
  if (is_text(type[i]))
    printf("extern void %s();\n",name[i]);
  else
    printf("extern char %s;\n",name[i]);
}

function print_strtab_begin() {
  printf("static char const strtab[] = {\n");
}

function print_strtab(i) {
  printf("/*%04d*/  %s0,\n",i,gensub(/(.)/,"'\\1',","g",name[i]));
}

function print_strtab_end() {
  printf("};\n\n");
}

function print_item_begin(){
  printf("#define NO_FUNCTION_PROTOTYPES\n");
  printf("#include <libstab.h>\n\n");
  printf("static int const stab_count = %d;\n", num_entries);
  printf("static stab_entry const stab[%d] = {\n", num_entries);
}

function print_item(i) {
  local_next = "0L";

  if(is_text(type[i]))
    address = name[i];
  else
    address = sprintf("&%s",name[i]);

  printf("/*%04d*/ {&strtab[%d],%s,%s,%s,-1,-1",
  	  i,
	  strtab_index,
	  address,
	  get_type(type[i]),
	  local_next);

  printf("},\n");
  
  strtab_index+=length(name[i])+1;
}

function print_item_end(){
  printf("};\n\n");
}

/:/ { next; }
(NF==0) { next; }

/ [DT] / {
    local_name = $3;
    local_type = $2;

    if(!is_weird_symbol(local_name) && 
       get_type(local_type)!="BAD"){
	type[num_entries] = local_type;
	name[num_entries] = local_name;
	num_entries++;
    }
}

END{
  # for each symbol, print the prototype:
  for(i=0;i<num_entries;i++)
    print_prototype(i);

  # for each symbol, print the name in the strtab:
  print_strtab_begin();
  for(i=0;i<num_entries;i++)
      print_strtab(i);
  print_strtab_end();

  # for each symbol, print the itemRec structure in stab:
  print_item_begin();
  for(i=0;i<num_entries;i++)
      print_item(i);
  print_item_end();

  printf("stab_t *%s = 0;\n", system_symbol_table);
  printf("stab_t *stab_create(int num_buckets);\n");
  printf("void %s_init(){\n", system_symbol_table);
  printf("  int i;\n");
  printf("  %s = stab_create(%d);\n",system_symbol_table,num_buckets);
  printf("  for(i=0;i<stab_count;i++)\n");
  printf("    stab_insert(%s,stab[i].name,stab[i].address,stab[i].type,-1,-1);\n",system_symbol_table);
  printf("  printf(\"%s_init() - %%d symbols added\\n\",i);\n", system_symbol_table);
  printf("}\n");
}
